import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


class MacdProcess(ProcessMethod):
    """
    macdを基準として今後上昇するかどうかをスコアで返す。
    値が大きければその傾向が高いことを表している。
    最小値は0で、最大値は無限大である。
    """

    short_term: int = 12
    long_term: int = 26
    macd_span: int = 9
    method_name: str = "macd"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        # histogramが図として表現されるMACDの値
        df = df.assign(
            # MACDの計算
            ema_short=lambda x: x["close"].ewm(span=self.short_term).mean(),
            ema_long=lambda x: x["close"].ewm(span=self.long_term).mean(),
            macd=lambda x: x["ema_short"] - x["ema_long"],
            signal=lambda x: x["macd"].ewm(span=self.macd_span).mean(),
            # ヒストグラム値
            histogram=lambda x: x["macd"] - x["signal"],
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        # 正負が交差した点
        df = df.join(self._cross(df["histogram"]))
        df = df.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df

    def _processed_columns(self) -> list:
        return ["ema_short", "ema_long", "signal", "macd", "histogram"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return {"signal": df_p["signal"].tail(3).mean(), "histogram": df_p["histogram"].tail(3).mean()}


class MacdVisualize(VisualizeMethod):
    """
    macdを基準として今後上昇するかどうかをスコアで返す。
    値が大きければその傾向が高いことを表している。
    最小値は0で、最大値は無限大である。
    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "macd", "color": "", "label": "macd", "plot": "plot"},
            {"df_key": "signal", "color": "", "label": "signal", "plot": "plot"},
            {"df_key": "histogram", "color": "", "label": "histogram", "plot": "bar"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "lower"}


macd = Method.of(process_method=MacdProcess(), visualize_method=MacdVisualize())



import math

import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


class StochasticsProcess(ProcessMethod):
    """
    買いのシグナルを計算で求める

    * %K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
    * %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時

    See Also:
        * https://www.moneypartners.co.jp/support/tech/sct.html

    """

    method_name: str = "stochastics"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["K"] = self._fast_stochastic_k(df_["close"], df_["low"], df_["high"], 9)
        df_["D"] = self._fast_stochastic_d(df_["K"])
        df_["SD"] = self._slow_stochastic_d(df_["D"])
        return df_

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        買いと売りに関する指標を算出する
        """
        df = df.assign(
            shift_K=lambda x: x["K"].shift(1),
            shift_D=lambda x: x["D"].shift(1),
            shift_SD=lambda x: x["SD"].shift(1),
        ).fillna(0)

        # 複数引数は関数を利用することで吸収
        df["buy_signal"] = df.apply(self._buy_signal_index_internal, axis=1)
        df["sell_signal"] = df.apply(self._sell_signal_index_internal, axis=1)
        return df

    @staticmethod
    def _fast_stochastic_k(close, low, high, n):
        return (
            (close - low.rolling(window=n, center=False).min())
            / (high.rolling(window=n, center=False).max() - low.rolling(window=n, center=False).min())
        ) * 100

    @staticmethod
    def _fast_stochastic_d(stochastic_k):
        # ストキャスの%Dを計算（%Kの3日SMA）
        return stochastic_k.rolling(window=3, center=False).mean()

    @staticmethod
    def _slow_stochastic_d(stochastic_d):
        # ストキャスの%SDを計算（%Dの3日SMA）
        return stochastic_d.rolling(window=3, center=False).mean()

    @staticmethod
    def _buy_signal_index_internal(x: pd.Series) -> float:
        return StochasticsProcess._buy_signal_index(x["K"], x["D"], x["SD"], x["shift_K"], x["shift_D"], x["shift_SD"])

    @staticmethod
    def _buy_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k > 30) | (current_d > 30) | (current_sd > 30):
            return 0

        # %K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
        if current_k < 20 and current_d < 20:
            if (prev_d > prev_k) and (current_d < current_k):
                return current_k - current_d

        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d < 20 and current_sd < 20:
            if (prev_sd > prev_d) and (current_sd < current_d):
                return current_d - current_sd
        return 1 / math.exp(
            math.pow(current_k - 20, 2) / 100 + math.pow(current_d - 20, 2) / 100 + math.pow(current_sd - 20, 2) / 100
        )

    @staticmethod
    def _sell_signal_index_internal(x: pd.Series) -> float:
        return StochasticsProcess._sell_signal_index(x["K"], x["D"], x["SD"], x["shift_K"], x["shift_D"], x["shift_SD"])

    @staticmethod
    def _sell_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k < 70) | (current_d < 70) | (current_sd < 70):
            return 0

        # %K・%D共に80％以上の時に、%Kが%Dを上から下抜いた時
        if current_k > 80 and current_d > 80:
            if (prev_d < prev_k) and (current_d > current_k):
                return current_d - current_k

        # %D・スロー%D共に80％以上の時に、%Dがスロー%Dを上から下抜いた時
        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d > 80 and current_sd > 80:
            if (prev_sd < prev_d) and (current_sd > current_d):
                return current_d - current_sd
        return 1 / math.exp(
            math.pow(current_k - 20, 2) / 100 + math.pow(current_d - 20, 2) / 100 + math.pow(current_sd - 20, 2) / 100
        )

    def _processed_columns(self) -> list:
        return ["K", "D", "SD"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return {
            "stochastics_k": df_p["K"].tail(3).mean(),
            "stochastics_d": df_p["D"].tail(3).mean(),
            "stochastics_sd": df_p["SD"].tail(3).mean(),
        }


class StochasticsVisualize(VisualizeMethod):
    """
    買いのシグナルを計算で求める

    * %K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
    * %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時

    See Also:
        * https://www.moneypartners.co.jp/support/tech/sct.html

    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "K", "color": "#dc143c", "label": "%K"},
            {"df_key": "D", "color": "#ffa500", "label": "%D"},
            {"df_key": "SD", "color": "#1e90ff", "label": "%SD"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "lower"}


stochastics = Method.of(process_method=StochasticsProcess(), visualize_method=StochasticsVisualize())



import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


class SmaProcess(ProcessMethod):
    """
    SMAを計算する
    """

    short_term: int = 5
    medium_term: int = 21
    long_term: int = 70
    method_name: str = "sma"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            sma_short=df["close"].rolling(self.short_term).mean(),
            sma_medium=df["close"].rolling(self.medium_term).mean(),
            sma_long=df["close"].rolling(self.long_term).mean(),
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df["diff"] = df.apply(lambda x: x["sma_long"] - x["sma_short"], axis=1)
        # 正負が交差した点
        df = df.join(self._cross(df["diff"]))
        df = df.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df

    def _processed_columns(self) -> list:
        return ["sma_long", "sma_medium", "sma_short"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        # difference from close
        df_p["sma_short_diff"] = (df_p["sma_short"] - df_x["close"]) / df_p["sma_short"]
        df_p["sma_medium_diff"] = (df_p["sma_medium"] - df_x["close"]) / df_p["sma_medium"]
        df_p["sma_long_diff"] = (df_p["sma_long"] - df_x["close"]) / df_p["sma_long"]
        # difference from sma_long
        df_p["sma_long_short"] = (df_p["sma_long"] - df_p["sma_short"]) / df_p["sma_long"]
        df_p["sma_long_medium"] = (df_p["sma_long"] - df_p["sma_medium"]) / df_p["sma_long"]
        return {
            "sma_short_diff": df_p["sma_short_diff"].tail(3).mean(),
            "sma_medium_diff": df_p["sma_medium_diff"].tail(3).mean(),
            "sma_long_diff": df_p["sma_long_diff"].tail(3).mean(),
            "sma_long_short": df_p["sma_long_short"].tail(3).mean(),
            "sma_long_medium": df_p["sma_long_medium"].tail(3).mean(),
        }


class SmaVisualize(VisualizeMethod):
    """
    SMAを計算する
    """

    short_term: int = 5
    medium_term: int = 21
    long_term: int = 70

    def _color_mapping(self) -> list:
        return [
            {"df_key": "sma_long", "color": "#dc143c", "label": f"sma({self.long_term})"},
            {"df_key": "sma_medium", "color": "#ffa500", "label": f"sma({self.medium_term})"},
            {"df_key": "sma_short", "color": "#1e90ff", "label": f"sma({self.short_term})"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "in"}


sma = Method.of(process_method=SmaProcess(), visualize_method=SmaVisualize())



from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Optional

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from mplfinance.original_flavor import candlestick_ohlc
from pydantic import BaseModel, ConfigDict

from kabutobashi.domain.errors import KabutobashiEntityError
from kabutobashi.domain.values import StockDataProcessed, StockDataVisualized

__all__ = ["ProcessMethod", "VisualizeMethod", "Method", "MethodType"]


class MethodType(Enum):
    TECHNICAL_ANALYSIS = auto()
    PARAMETERIZE = auto()
    CROSS_SECTION = auto()
    SIGNAL_PROCESSING = auto()


class ProcessMethod(ABC, BaseModel):
    """
    株のテクニカル分析に関するメソッドを提供するクラス

    Examples:
        >>> import pandas as pd
        >>> import kabutobashi as kb
        >>> stock_df: pd.DataFrame = pd.DataFrame("path_to_stock_data")
        # get sma-based-analysis
        >>> sma_df = stock_df.pipe(kb.sma)
        # get sma-base-buy or sell signal
        >>> sma_signal = stock_df.pipe(kb.sma, impact="true", influence=5, tail=5)
        # get macd-based-analysis
        >>> macd_df = stock_df.pipe(kb.macd)
        # get macd-base-buy or sell signal
        >>> sma_signal = stock_df.pipe(kb.macd, impact="true", influence=5, tail=5)
    """

    # 名前
    method_name: str
    # 種類:
    method_type: MethodType
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def process(self, df: pd.DataFrame, influence: int = 2, tail: int = 5) -> "StockDataProcessed":
        code_list = list(set(df["code"].values))
        if len(code_list) != 1:
            raise KabutobashiEntityError
        # 日時
        start_at = list(df["dt"])[0]
        end_at = list(df["dt"])[-1]

        required_columns = [
            "dt",
            "open",
            "close",
            "high",
            "low",
            "buy_signal",
            "sell_signal",
        ] + self.processed_columns()
        applied_df = self.apply(df=df)

        signal_df = self.signal(df=applied_df)
        params = self.parameterize(df_x=applied_df, df_p=signal_df)
        params.update({self.method_name: self._get_impact(df=signal_df, influence=influence, tail=tail)})
        return StockDataProcessed(
            code=code_list[0],
            df=signal_df,
            start_at=start_at,
            end_at=end_at,
            df_required_columns=required_columns,
            applied_method_name=self.method_name,
            parameters=params,
        )

    def apply(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        テクニカル分析の手法

        Args:
            df: 株の情報を含むDataFrame

        Returns:
            各分析手法の結果を付与したDataFrame
        """
        return self._apply(df=df)

    @abstractmethod
    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("please implement your code")  # pragma: no cover

    def processed_columns(self) -> list:
        return self._processed_columns()

    @abstractmethod
    def _processed_columns(self) -> list:
        """
        各メソッドで計算時に出力されるカラムを明示する

        Returns:

        """
        raise NotImplementedError("please implement your code")  # pragma: no cover

    def signal(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        テクニカル分析の手法の結果により、買いと売りのタイミングを計算する

        Args:
            df: 株の情報を含むDataFrame

        Returns:

        """
        return self._signal(df=df)

    @abstractmethod
    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("please implement your code")  # pragma: no cover

    @staticmethod
    def _cross(_s: pd.Series, to_plus_name=None, to_minus_name=None) -> pd.DataFrame:
        """
        0を基準としてプラスかマイナスのどちらかに振れたかを判断する関数

        Args:
            _s: 対象のpd.Series
            to_plus_name: 上抜けた場合のカラムの名前
            to_minus_name: 下抜けた場合のカラムの名前
        """
        # shorten variable name
        col = "original"
        shifted = "shifted"

        # shiftしたDataFrameの作成
        shift_s = _s.shift(1)
        df = pd.DataFrame({col: _s, shifted: shift_s})

        # 正負が交差した点
        df = df.assign(
            is_cross=df.apply(lambda x: 1 if x[col] * x[shifted] < 0 else 0, axis=1),
            is_higher=df.apply(lambda x: 1 if x[col] > x[shifted] else 0, axis=1),
            is_lower=df.apply(lambda x: 1 if x[col] < x[shifted] else 0, axis=1),
        )

        # 上抜けか下抜けかを判断している
        df = df.assign(to_plus=df["is_cross"] * df["is_higher"], to_minus=df["is_cross"] * df["is_lower"])
        if to_plus_name is not None:
            df = df.rename(columns={"to_plus": to_plus_name})
        if to_minus_name is not None:
            df = df.rename(columns={"to_minus": to_minus_name})
        return df

    @staticmethod
    def _trend(_s: pd.Series) -> pd.Series:
        """
        ある系列_sのトレンドを計算する。
        差分のrolling_sumを返す
        """
        # shorten variable name
        col = "original"
        shifted = "shifted"

        # shiftしたDataFrameの作成
        shift_s = _s.shift(1)
        df = pd.DataFrame({col: _s, shifted: shift_s})
        df["diff"] = df["original"] - df["shifted"]
        df["diff_rolling_sum"] = df["diff"].rolling(5).sum()
        return df["diff_rolling_sum"]

    def parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return self._parameterize(df_x=df_x, df_p=df_p)

    @abstractmethod
    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        raise NotImplementedError("please implement your code")  # pragma: no cover

    @staticmethod
    def _get_impact(df: pd.DataFrame, influence: int, tail: int) -> float:
        """
        売りと買いのシグナルの余波の合計値を返す。

        Args:
            df:
            influence:
            tail:

        Returns:
            [-1,1]の値をとる。-1: 売り、1: 買いを表す
        """
        columns = df.columns
        if "buy_signal" not in columns:
            return 0
        if "sell_signal" not in columns:
            return 0

        df["buy_impact"] = df["buy_signal"].ewm(span=influence).mean()
        df["sell_impact"] = df["sell_signal"].ewm(span=influence).mean()
        buy_impact_index = df["buy_impact"].iloc[-tail:].sum()
        sell_impact_index = df["sell_impact"].iloc[-tail:].sum()
        return round(buy_impact_index - sell_impact_index, 5)


class VisualizeMethod(ABC, BaseModel):
    def color_mapping(self) -> list:
        return self._color_mapping()

    @abstractmethod
    def _color_mapping(self) -> list:
        raise NotImplementedError("please implement your code")  # pragma: no cover

    def visualize_option(self) -> dict:
        return self._visualize_option()

    @abstractmethod
    def _visualize_option(self) -> dict:
        raise NotImplementedError("please implement your code")  # pragma: no cover

    @staticmethod
    def _add_ax_candlestick(ax, _df: pd.DataFrame):
        # datetime -> float
        time_series = mdates.date2num(_df["dt"])
        data = _df[["open", "high", "low", "close"]].values.T
        # data
        ohlc = np.vstack((time_series, data)).T
        candlestick_ohlc(ax, ohlc, width=0.7, colorup="g", colordown="r")

    def visualize(self, size_ratio: int, df: pd.DataFrame) -> StockDataVisualized:
        return StockDataVisualized(size_ratio=size_ratio, fig=self._visualize(size_ratio=size_ratio, df=df))

    def _visualize(self, size_ratio: int, df: pd.DataFrame):
        """
        Visualize Stock Data.

        Args:
            size_ratio: determine the size of the graph, default 2.

        Returns:
            Figure
        """

        def _n_rows() -> int:
            lower_nums = 1 if self.visualize_option()["position"] == "lower" else 0
            return 1 + lower_nums

        n_rows = _n_rows()

        def _gridspec_kw() -> dict:
            if n_rows == 1:
                return {"height_ratios": [3]}
            return {"height_ratios": [3] + [1] * (n_rows - 1)}

        gridspec_kw = _gridspec_kw()
        fig, axs = plt.subplots(
            nrows=n_rows, ncols=1, figsize=(6 * size_ratio, 5 * size_ratio), gridspec_kw=gridspec_kw
        )
        # auto-formatting x-axis
        fig.autofmt_xdate()

        # set candlestick base
        base_df = df[["dt", "open", "close", "high", "low"]]
        if n_rows == 1:
            base_axs = axs
        else:
            base_axs = axs[0]
        self._add_ax_candlestick(base_axs, base_df)

        ax_idx = 1
        # plots
        position = self.visualize_option()["position"]
        time_series = mdates.date2num(base_df["dt"])
        mapping = self.color_mapping()
        if position == "in":
            for m in mapping:
                df_key = m["df_key"]
                color = m["color"]
                label = m["label"]
                base_axs.plot(time_series, df[df_key], label=label)
            # display labels
            base_axs.legend(loc="best")
        elif position == "lower":
            for m in mapping:
                df_key = m["df_key"]
                color = m["color"]
                label = m["label"]
                plot = m.get("plot", "plot")
                if plot == "plot":
                    # type FloatingArray is no accepted ...
                    # so `df[df_key].astype(float)`
                    axs[ax_idx].plot(time_series, df[df_key].astype(float), label=label)
                elif plot == "bar":
                    axs[ax_idx].bar(time_series, df[df_key], label=label)
            # display labels
            axs[ax_idx].legend(loc="best")
            # lower
            ax_idx += 1
        elif position == "-":
            # technical_analysis以外のmethodが入っている場合
            pass
        else:
            raise KabutobashiEntityError()

        return fig


class Method(BaseModel):
    process_method: ProcessMethod
    visualize_method: Optional[VisualizeMethod]

    @staticmethod
    def of(process_method: ProcessMethod, visualize_method: Optional[VisualizeMethod]):
        return Method(process_method=process_method, visualize_method=visualize_method)



"""
define structures of the stock-data,
when processing Methods like SMA, MCAD,
and when estimating stock-code which is to rise in the next day or so on.

- Used for ``crawling``

  - value_object.StockIpo

- define data-structure: ``basement``

  - entity.Stock
  - aggregates.StockCodeSingleAggregate

- initial step to analyze:  ``processed``

  - value_object.StockDataProcessed

- second step to analyze:  ``estimated``

  - value_object.StockDataEstimated
"""



import pytest

import kabutobashi as kb


@pytest.fixture(scope="module", autouse=True)
def stock_agg() -> kb.StockCodeSingleAggregate:
    df = kb.example()
    yield kb.StockCodeSingleAggregate.of(entity=df, code="1375")


def test_example_data(stock_agg):
    columns = stock_agg.stock.to_df().columns
    assert "dt" in columns
    assert "open" in columns
    assert "close" in columns
    assert "high" in columns
    assert "low" in columns


def test_analysis_with_sma(stock_agg):
    processed = stock_agg.with_processed([kb.sma])
    columns = processed.processed_list[0].df.columns
    assert "sma_short" in columns
    assert "sma_medium" in columns
    assert "sma_long" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "sma"


def test_analysis_with_macd(stock_agg):
    processed = stock_agg.with_processed([kb.macd])
    columns = processed.processed_list[0].df.columns
    assert "ema_short" in columns
    assert "ema_long" in columns
    assert "signal" in columns
    assert "macd" in columns
    assert "histogram" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "macd"


def test_analysis_with_stochastics(stock_agg):
    processed = stock_agg.with_processed([kb.stochastics])
    columns = processed.processed_list[0].df.columns
    assert "K" in columns
    assert "D" in columns
    assert "SD" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "stochastics"


def test_analysis_with_adx(stock_agg):
    processed = stock_agg.with_processed([kb.adx])
    columns = processed.processed_list[0].df.columns
    assert "plus_di" in columns
    assert "minus_di" in columns
    assert "DX" in columns
    assert "ADX" in columns
    assert "ADXR" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "adx"


@pytest.mark.skip(reason="buy_signal and sell_signal is not implemented")
def test_analysis_with_ichimoku(stock_agg):
    processed = stock_agg.with_processed([kb.ichimoku])
    columns = processed.processed_list[0].df.columns
    assert "line_change" in columns
    assert "line_base" in columns
    assert "proceding_span_1" in columns
    assert "proceding_span_2" in columns
    assert "delayed_span" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "ichimoku"


def test_analysis_with_momentum(stock_agg):
    processed = stock_agg.with_processed([kb.momentum])
    columns = processed.processed_list[0].df.columns
    assert "momentum" in columns
    assert "sma_momentum" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "momentum"


def test_analysis_with_psycho_logical(stock_agg):
    processed = stock_agg.with_processed([kb.psycho_logical])
    columns = processed.processed_list[0].df.columns
    assert "psycho_line" in columns
    assert "bought_too_much" in columns
    assert "sold_too_much" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "psycho_logical"


def test_analysis_with_bollinger_bands(stock_agg):
    processed = stock_agg.with_processed([kb.bollinger_bands])
    columns = processed.processed_list[0].df.columns
    assert "upper_2_sigma" in columns
    assert "lower_2_sigma" in columns
    assert "over_upper_continuity" in columns
    assert "over_lower_continuity" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "bollinger_bands"


def test_analysis_with_fitting(stock_agg):
    processed = stock_agg.with_processed([kb.fitting])
    columns = processed.processed_list[0].df.columns
    assert "linear_fitting" in columns
    assert "square_fitting" in columns
    assert "cube_fitting" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "fitting"


def test_analysis_with_basic(stock_agg):
    processed = stock_agg.with_processed([kb.basic])
    columns = processed.processed_list[0].df.columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns
    # method name
    assert processed.processed_list[0].applied_method_name == "basic"


def test_analysis_with_pct_change(stock_agg):
    processed = stock_agg.with_processed([kb.pct_change])
    # method name
    assert processed.processed_list[0].applied_method_name == "pct_change"


@pytest.mark.skip(reason="industry_categories is not prepared")
def test_analysis_with_industry_categories(stock_agg):
    processed = stock_agg.with_processed([kb.industry_categories])
    # method name
    assert processed.processed_list[0].applied_method_name == "industry_categories"


def test_analysis_with_volatility(stock_agg):
    processed = stock_agg.with_processed([kb.volatility])
    # method name
    assert processed.processed_list[0].applied_method_name == "volatility"


def test_get_impact_with(stock_agg):
    result_1 = stock_agg.with_processed(methods=[kb.sma])
    assert "sma" in [v.applied_method_name for v in result_1.processed_list]
    result_2 = stock_agg.with_processed(methods=[kb.sma, kb.macd])
    assert "sma" in [v.applied_method_name for v in result_2.processed_list]
    assert "macd" in [v.applied_method_name for v in result_2.processed_list]



from typing import Any, Dict, List

import matplotlib.pyplot as plt
import pandas as pd
from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic_core.core_schema import FieldValidationInfo

from kabutobashi.domain.errors import KabutobashiEntityError


class StockDataProcessed(BaseModel):
    """
    StockDataProcessed: ValueObject
    Holds data processed by singular-Method.
    """

    code: str
    start_at: str
    end_at: str
    applied_method_name: str
    df: pd.DataFrame = Field(repr=False)
    df_required_columns: List[str] = Field(repr=False)
    parameters: Dict[str, Any]
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    @field_validator("df")
    def _validate_df(cls, v, info: FieldValidationInfo):
        df_required_columns = info.data["df_required_columns"]
        df_columns = v.columns
        if not all([c in df_columns for c in df_required_columns]):
            raise KabutobashiEntityError()


class StockDataVisualized(BaseModel):
    """
    StockDataVisualized: ValueObject
    Used to visualize.
    """

    fig: plt.Figure
    size_ratio: int
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    @field_validator("fig")
    def _validate_fig(cls, v):
        if type(v) is not plt.Figure:
            raise KabutobashiEntityError()


class StockDataEstimated(BaseModel):
    """
    StockDataEstimated: ValueObject
    """

    code: str
    estimated_value: float
    estimate_filter_name: str

    def weighted_estimated_value(self, weights: dict) -> float:
        weight = weights.get(self.estimate_filter_name, 1)
        return weight * self.estimated_value



class KabutobashiBaseError(Exception):
    pass


class KabutobashiPageError(KabutobashiBaseError):
    """
    Crawlが失敗したときに返すエラー
    """

    def __init__(self, url: str = ""):
        self.url = url

    def __str__(self):
        return f"error occurred when crawling [{self.url}]"


class TagNotFoundError(KabutobashiPageError):
    """
    crawlしたいページに対象のtagがない場合に返すエラー
    """

    def __init__(self, tag):
        super().__init__(url="")
        self.tag = tag

    def __str__(self):
        return f"tag [{self.tag}] not found"


class KabutobashiMethodError(KabutobashiBaseError):
    pass


class KabutobashiVisualizeError(KabutobashiBaseError):
    pass


class KabutobashiEntityError(KabutobashiBaseError):
    pass



from kabutobashi.domain.services.decode_html import StockInfoMinkabuTopHtmlDecoder, StockIpoHtmlDecoder
from kabutobashi.domain.values import DecodeHtmlPageStockInfoMinkabuTop
from kabutobashi.infrastructure.repository import StockInfoHtmlPageRepository, StockIpoHtmlPageRepository


def test_crawl_page_detail():
    html_repo = StockInfoHtmlPageRepository(code=4395)
    html_page = html_repo.read()
    result = StockInfoMinkabuTopHtmlDecoder().decode_to_dict(html_page=html_page)
    assert result is not None
    assert type(result) is dict

    result_object = StockInfoMinkabuTopHtmlDecoder().decode_to_object(html_page=html_page)
    assert result_object is not None
    assert type(result_object) is DecodeHtmlPageStockInfoMinkabuTop


def test_crawl_ipo_list():
    html_page = StockIpoHtmlPageRepository(year="2019").read()
    result = StockIpoHtmlDecoder().decode_to_dict(html_page=html_page)
    assert result is not None
    assert type(result) is dict



import os

import pandas as pd

PARENT_PATH = os.path.abspath(os.path.dirname(__file__))
SOURCE_PATH = os.path.abspath(os.path.dirname(PARENT_PATH))
DATA_PATH = f"{SOURCE_PATH}/data"


def example() -> pd.DataFrame:
    """

    Examples:
        >>> import kabutobashi as kb
        >>> records = kb.example()
        >>> agg = kb.StockCodeSingleAggregate.of(entity=records, code="1375")
        >>> processed = agg.with_processed(kb.methods)
    """
    file_name = "example.csv.gz"
    return pd.read_csv(f"{DATA_PATH}/{file_name}")



from .decoded_html_pages import (
    DecodedHtmlPage,
    DecodeHtmlPageStockInfoMinkabuTop,
    DecodeHtmlPageStockInfoMultipleDays,
    DecodeHtmlPageStockIpo,
)
from .raw_html_pages import (
    IHtmlPageRepository,
    RawHtmlPage,
    RawHtmlPageStockInfo,
    RawHtmlPageStockInfoMultipleDaysMain,
    RawHtmlPageStockInfoMultipleDaysSub,
    RawHtmlPageStockIpo,
)
from .stock_data import StockDataEstimated, StockDataProcessed, StockDataVisualized
from .user_agent import UserAgent



from abc import ABC, abstractmethod

import pandas as pd


class IDictSerialize(ABC):
    @abstractmethod
    def to_dict(self) -> dict:
        raise NotImplementedError()  # pragma: no cover

    @staticmethod
    @abstractmethod
    def from_dict(data: dict):
        raise NotImplementedError()  # pragma: no cover


class IDfSerialize(ABC):
    @abstractmethod
    def to_df(self) -> pd.DataFrame:
        raise NotImplementedError()  # pragma: no cover

    @staticmethod
    @abstractmethod
    def from_df(data: pd.DataFrame):
        raise NotImplementedError()  # pragma: no cover



from kabutobashi import Stock
from kabutobashi.application import crawl_info, crawl_info_multiple, crawl_ipo


class TestStockInfo:
    def test_pass(self):
        res = crawl_info(code="9101")
        assert type(res) is Stock


class TestStockInfoMultiple:
    def test_pass(self):
        res = crawl_info_multiple(code="9101")
        assert type(res) is Stock


class TestIpo:
    def test_pass(self):
        res = crawl_ipo(year="2022")
        assert type(res) is list
        assert all([(type(v) is Stock) for v in res])



import pytest

import kabutobashi as kb
from kabutobashi.domain.errors import KabutobashiEntityError
from kabutobashi.utilities import convert_float, convert_int, replace


def test_workday():
    date_list = kb.get_past_n_days("2020-03-31", 30)
    assert len(date_list) == 30
    # 祝日が除かれていることを確認
    assert "2020-03-20" not in date_list
    # 土日が除かれていることを確認
    assert "2020-03-07" not in date_list
    assert "2020-03-08" not in date_list
    assert "2020-03-14" not in date_list
    assert "2020-03-15" not in date_list
    assert "2020-03-21" not in date_list
    assert "2020-03-22" not in date_list
    assert "2020-03-28" not in date_list
    assert "2020-03-29" not in date_list


def test_replace():
    assert replace("-") == "0"
    assert replace("") == "0"
    assert replace("---") == "0"
    assert replace("1,000") == "1000"
    assert replace("1000円") == "1000"
    assert replace("1000株") == "1000"
    assert replace("1000倍") == "1000"
    assert replace("1,000倍株円") == "1000"


def test_convert_int():
    # pass
    assert convert_int("-") == 0
    assert convert_int("") == 0
    assert convert_int("---") == 0
    assert convert_int("1,000") == 1000
    assert convert_int("1000円") == 1000
    assert convert_int("1000株") == 1000
    assert convert_int("1000倍") == 1000
    assert convert_int("1,000倍株円") == 1000
    assert type(convert_int(float(0))) is int
    assert convert_int(float(0)) == 0

    # error
    with pytest.raises(KabutobashiEntityError):
        convert_int(input_value={})

    with pytest.raises(KabutobashiEntityError):
        convert_int(input_value="a")


def test_convert_float():
    with pytest.raises(KabutobashiEntityError):
        convert_float(input_value={})

    with pytest.raises(KabutobashiEntityError):
        convert_float(input_value="a")



import pandas as pd
import pydantic
import pytest

import kabutobashi as kb
from kabutobashi.domain.entity.stock import Market, StockBrand, StockPriceRecord, StockReferenceIndicator
from kabutobashi.domain.errors import KabutobashiEntityError


@pytest.fixture(scope="module", autouse=True)
def entity() -> pd.DataFrame:
    yield kb.example()


class TestMarket:
    def test_get(self):
        assert Market.get(target="") is Market.NONE
        assert Market.get(target="東証プライ") is Market.NONE
        assert Market.get(target="東証スタンダー") is Market.NONE
        assert Market.get(target="東証グロー") is Market.NONE
        assert Market.get(target=None) is Market.NONE
        assert Market.get(target="東証プライム") is Market.TOKYO_STOCK_EXCHANGE_PRIME
        assert Market.get(target="東証スタンダード") is Market.TOKYO_STOCK_EXCHANGE_STANDARD
        assert Market.get(target="東証グロース") is Market.TOKYO_STOCK_EXCHANGE_GROWTH


class TestStockBrand:
    def test_error_init(self):
        # type-error
        with pytest.raises(pydantic.ValidationError):
            _ = StockBrand(
                id=None,
                code="1234",
                market="東証",
                name="例",
                unit="",
                market_capitalization="",
                industry_type="",
                issued_shares="",
            )

    def test_init(self):
        brand = StockBrand(
            id=None,
            code="1234",
            market="東証",
            name="例",
            unit=100,
            market_capitalization="",
            industry_type="",
            issued_shares="",
        )
        brand_dict = brand.to_dict()
        assert type(brand_dict) == dict
        assert brand_dict["code"] == "1234"
        assert brand_dict["market"] == "東証"
        assert brand_dict["name"] == "例"
        assert brand_dict["unit"] == 100

    def test_merge_error(self):
        brand1 = StockBrand(
            id=None,
            code="1234",
            market="東証",
            name="例",
            unit=100,
            market_capitalization="",
            industry_type="",
            issued_shares="",
        )
        brand2 = StockBrand(
            id=None,
            code="1235",
            market="東証",
            name="例",
            unit=100,
            market_capitalization="",
            industry_type="",
            issued_shares="",
        )
        with pytest.raises(KabutobashiEntityError):
            _ = brand1 + brand2
        with pytest.raises(KabutobashiEntityError):
            _ = brand1 + ""

    def test_merge_pass(self):
        brand1 = StockBrand(
            id=None,
            code="1234",
            market="東証",
            name="例",
            unit=100,
            market_capitalization=None,
            industry_type=None,
            issued_shares=None,
        )
        merged = brand1 + None
        assert merged.code == "1234"

        brand2 = StockBrand(
            id=None,
            code="1234",
            market="東証",
            name=None,
            unit=None,
            market_capitalization="時価総額",
            industry_type="業種",
            issued_shares="発行済み株式数",
        )
        merged = brand1 + brand2
        assert merged.code == "1234"
        assert merged.name == "例"
        assert merged.market == "東証"
        assert merged.unit == 100
        assert merged.market_capitalization == "時価総額"
        assert merged.industry_type == "業種"
        assert merged.issued_shares == "発行済み株式数"
        merged = brand2 + brand1
        assert merged.code == "1234"
        assert merged.name == "例"
        assert merged.market == "東証"
        assert merged.unit == 100
        assert merged.market_capitalization == "時価総額"
        assert merged.industry_type == "業種"
        assert merged.issued_shares == "発行済み株式数"


class TestStockRecord:
    def test_pass_init(self):
        _ = StockPriceRecord(
            id=None,
            code="1234",
            open=0,
            high=0,
            low=0,
            close=0,
            volume=0,
            dt="",
        )

    def test_eq_test(self):
        price1 = StockPriceRecord(
            id=None,
            code="1234",
            open=0,
            high=0,
            low=0,
            close=0,
            volume=0,
            dt="2022-01-01",
        )
        price2 = StockPriceRecord(
            id=None,
            code="1234",
            open=1,
            high=1,
            low=1,
            close=1,
            volume=1,
            dt="2022-01-01",
        )
        price3 = StockPriceRecord(
            id=None,
            code="1235",
            open=2,
            high=2,
            low=2,
            close=2,
            volume=2,
            dt="2022-01-01",
        )
        price4 = StockPriceRecord(
            id=None,
            code="1234",
            open=3,
            high=3,
            low=3,
            close=3,
            volume=3,
            dt="2022-01-02",
        )
        assert price1 == price2
        assert price1 != price3
        assert price2 != price3
        assert price1 != price4
        assert price2 != price4
        assert price3 != price4
        assert price1 != ""
        assert price2 != ""
        assert price3 != ""
        assert price4 != ""


class TestStockReferenceIndicator:
    def test_pass_init(self):
        _ = StockReferenceIndicator(id=0, code="1234", dt="2023-01-01", pbr=0, psr=0, per=0)
        _ = StockReferenceIndicator.from_dict(
            {"id": 0, "code": "1234", "dt": "2023-01-01", "pbr": 0, "psr": 0, "per": 0}
        )

    def test_error_merge(self):
        reference1 = StockReferenceIndicator(id=0, code="1234", dt="2023-01-01", pbr=0, psr=0, per=0)
        reference2 = StockReferenceIndicator(id=0, code="1235", dt="2023-01-01", pbr=0, psr=0, per=0)
        with pytest.raises(KabutobashiEntityError):
            _ = reference1 + reference2
        with pytest.raises(KabutobashiEntityError):
            _ = reference1 + ""

    def test_pass_merge(self):
        reference1 = StockReferenceIndicator(id=0, code="1234", dt="2023-01-01", pbr=None, psr=None, per=None)
        reference2 = StockReferenceIndicator(id=0, code="1234", dt="2023-01-01", pbr=1, psr=2, per=3)
        merge = reference1 + None
        assert merge.code == "1234"
        merge = reference1 + reference2
        assert merge.code == "1234"
        assert merge.dt == "2023-01-01"
        assert merge.pbr == 1
        assert merge.psr == 2
        assert merge.per == 3


class TestStockIpo:
    def test_pass_init(self):
        _ = kb.DecodeHtmlPageStockIpo(
            code="", manager="", stock_listing_at="", public_offering=0, evaluation="", initial_price=0
        )


class TestStock:
    def test_init_error(self):
        brand = StockBrand(
            id=None,
            code="1234",
            market="東証",
            name="例",
            unit=100,
            market_capitalization=None,
            industry_type=None,
            issued_shares=None,
        )
        brand_e = StockBrand(
            id=None,
            code="1235",
            market="東証",
            name="例",
            unit=100,
            market_capitalization=None,
            industry_type=None,
            issued_shares=None,
        )
        price = StockPriceRecord(
            id=None,
            code="1234",
            open=0,
            high=0,
            low=0,
            close=0,
            volume=0,
            dt="2022-01-01",
        )
        price_e = StockPriceRecord(
            id=None,
            code="1235",
            open=0,
            high=0,
            low=0,
            close=0,
            volume=0,
            dt="2022-01-01",
        )
        reference = StockReferenceIndicator(id=0, code="1234", dt="2023-01-01", pbr=None, psr=None, per=None)
        reference_e = StockReferenceIndicator(id=0, code="1235", dt="2023-01-01", pbr=None, psr=None, per=None)
        with pytest.raises(KabutobashiEntityError):
            _ = kb.Stock(code="1234", brand=brand_e, daily_price_records=[price], reference_indicator=reference)
        with pytest.raises(KabutobashiEntityError):
            _ = kb.Stock(code="1234", brand=brand, daily_price_records=[price_e], reference_indicator=reference)
        with pytest.raises(KabutobashiEntityError):
            _ = kb.Stock(code="1234", brand=brand, daily_price_records=[price, price_e], reference_indicator=reference)
        with pytest.raises(KabutobashiEntityError):
            _ = kb.Stock(code="1234", brand=brand, daily_price_records=[price], reference_indicator=reference_e)

    def test_df_error(self, entity: pd.DataFrame):
        with pytest.raises(KabutobashiEntityError):
            _ = kb.Stock.from_df(data=entity)

    def test_df_pass(self, entity: pd.DataFrame):
        stock = kb.Stock.from_df(data=entity[entity["code"] == 1375])
        stock_df = stock.to_df(add_brand=True)
        stock_df_columns = stock_df.columns
        assert "industry_type" in stock_df_columns
        assert "market" in stock_df_columns
        assert "name" in stock_df_columns
        assert "is_delisting" in stock_df_columns
        assert "pbr" in stock_df_columns
        assert "per" in stock_df_columns
        assert "psr" in stock_df_columns

    def test_merge_error(self):
        brand = StockBrand(
            id=None,
            code="1234",
            market="東証",
            name="例",
            unit=100,
            market_capitalization=None,
            industry_type=None,
            issued_shares=None,
        )
        price = StockPriceRecord(
            id=None,
            code="1234",
            open=0,
            high=0,
            low=0,
            close=0,
            volume=0,
            dt="2022-01-01",
        )
        reference = StockReferenceIndicator(id=0, code="1234", dt="2023-01-01", pbr=None, psr=None, per=None)
        stock = kb.Stock(code="1234", brand=brand, daily_price_records=[price], reference_indicator=reference)
        with pytest.raises(KabutobashiEntityError):
            _ = stock + ""

        # 異なるcode
        brand_2 = StockBrand(
            id=None,
            code="1235",
            market="東証",
            name="例",
            unit=100,
            market_capitalization=None,
            industry_type=None,
            issued_shares=None,
        )
        price_2 = StockPriceRecord(
            id=None,
            code="1235",
            open=0,
            high=0,
            low=0,
            close=0,
            volume=0,
            dt="2022-01-01",
        )
        reference_2 = StockReferenceIndicator(id=0, code="1235", dt="2023-01-01", pbr=None, psr=None, per=None)
        stock_2 = kb.Stock(code="1235", brand=brand_2, daily_price_records=[price_2], reference_indicator=reference_2)
        with pytest.raises(KabutobashiEntityError):
            _ = stock + stock_2

    def test_merge_pass(self):
        brand = StockBrand(
            id=None,
            code="1234",
            market="東証",
            name="例",
            unit=100,
            market_capitalization=None,
            industry_type=None,
            issued_shares=None,
        )
        price = StockPriceRecord(
            id=None,
            code="1234",
            open=0,
            high=0,
            low=0,
            close=0,
            volume=0,
            dt="2022-01-01",
        )
        reference = StockReferenceIndicator(id=0, code="1234", dt="2023-01-01", pbr=None, psr=None, per=None)
        stock = kb.Stock(code="1234", brand=brand, daily_price_records=[price], reference_indicator=reference)
        merge = stock + None
        assert merge.code == "1234"

        merge = stock + stock
        assert merge.code == "1234"
        assert len(merge.daily_price_records) == 1

    def test_reduce_error(self, entity: pd.DataFrame):
        stock1 = kb.Stock.from_df(data=entity[entity["code"] == 1375])
        stock2 = kb.Stock.from_df(data=entity[entity["code"] == 1439])
        with pytest.raises(KabutobashiEntityError):
            kb.Stock.reduce(stocks=[stock1, stock2])

    def test_reduce_pass(self, entity: pd.DataFrame):
        filtered_entity = entity[entity["code"] == 1375]
        stock1 = kb.Stock.from_df(data=filtered_entity[:10])
        stock2 = kb.Stock.from_df(data=filtered_entity[10:])
        stock3 = kb.Stock.from_df(data=filtered_entity)

        assert len(kb.Stock.reduce(stocks=[stock1, stock2]).daily_price_records) == len(stock3.daily_price_records)


class TestStockSingleAggregate:
    def test_pass_init(self, entity: pd.DataFrame):
        stock = kb.Stock.from_df(data=entity[entity["code"] == 1375])
        _ = kb.StockCodeSingleAggregate.of(entity=stock)

    def test_error_init(self):
        with pytest.raises(KabutobashiEntityError):
            _ = kb.StockCodeSingleAggregate.of(entity="")

    def test_pass_methods(self, entity: pd.DataFrame):
        methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        agg = kb.StockCodeSingleAggregate.of(entity=entity, code="1375")
        estimated = agg.with_processed(methods=methods).with_estimated(stock_analysis=kb.stock_analysis)
        value = estimated.weighted_estimated_value({"fundamental": 1.0, "volume": 1.0})
        assert value != 0

        # check visualize single column
        data_visualized = agg.visualize(kb.sma)
        assert data_visualized.fig

        # check visualize multiple columns
        data_visualized = agg.visualize(kb.macd)
        assert data_visualized.fig

        data = estimated.to_dict()
        assert "code" in data
        assert "name" in data
        assert "market" in data
        assert "industry_type" in data
        assert "weighted_estimated_value" in data
        assert "estimate_name" in data

        with pytest.raises(NotImplementedError):
            kb.StockCodeSingleAggregate.from_dict(data={})

    def test_error_methods(self, entity: pd.DataFrame):
        methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        agg = kb.StockCodeSingleAggregate.of(entity=entity, code="1375")
        with pytest.raises(KabutobashiEntityError):
            _ = agg.with_processed(methods=[""])



import re
from datetime import datetime
from enum import Enum
from typing import List, Optional

import jpholiday
import pandas as pd
from pydantic import BaseModel, Field

from kabutobashi.domain.errors import KabutobashiEntityError
from kabutobashi.domain.serialize import IDfSerialize, IDictSerialize
from kabutobashi.utilities import convert_float, convert_int

__all__ = ["Market", "StockBrand", "StockPriceRecord", "StockReferenceIndicator", "Stock"]


class Market(Enum):
    TOKYO_STOCK_EXCHANGE_PRIME = ("東証プライム", "^(東証|東京証券取引所).*?(プライム).*?$")
    TOKYO_STOCK_EXCHANGE_STANDARD = ("東証スタンダード", "^(東証|東京証券取引所).*?(スタンダード).*?$")
    TOKYO_STOCK_EXCHANGE_GROWTH = ("東証グロース", "^(東証|東京証券取引所).*?(グロース).*?$")
    NONE = ("該当無し", "NONE")

    def __init__(self, market_name: str, regex: str):
        self.market_name = market_name
        self.regex = regex

    @staticmethod
    def get(target: Optional[str]):
        if target is None:
            return Market.NONE
        for v in list(Market):
            if re.match(v.regex, target):
                return v
        return Market.NONE


class StockBrand(BaseModel, IDictSerialize):
    """
    Model: Entity
    JP: 銘柄
    """

    id: int = Field(description="id")
    code: str = Field(description="銘柄コード")
    unit: Optional[int] = Field(description="単位")
    market: Optional[str] = Field(description="市場")
    name: Optional[str] = Field(description="銘柄名")
    industry_type: Optional[str] = Field(description="業種")
    market_capitalization: Optional[str] = Field(description="時価総額")
    issued_shares: Optional[str] = Field(description="発行済み株式")
    is_delisting: bool = Field(description="上場廃止")

    def __init__(
        self,
        code: str,
        id: Optional[int] = None,
        unit: Optional[int] = None,
        market: Optional[str] = None,
        name: Optional[str] = None,
        industry_type: Optional[str] = None,
        market_capitalization: Optional[str] = None,
        issued_shares: Optional[str] = None,
        is_delisting: bool = False,
    ):
        # code may "100.0"
        code = code.split(".")[0]

        super().__init__(
            id=0 if id is None else id,
            code=code,
            unit=unit,
            market=market,
            name=name,
            industry_type=industry_type,
            market_capitalization=market_capitalization,
            issued_shares=issued_shares,
            is_delisting=is_delisting,
        )

    @staticmethod
    def from_dict(data: dict) -> "StockBrand":
        code = str(data["code"]).split(".")[0]

        return StockBrand(
            id=data.get("id"),
            code=code,
            unit=convert_int(data.get("unit", 0)),
            market=Market.get(target=data.get("market")).market_name,
            name=data.get("name"),
            industry_type=data.get("industry_type"),
            market_capitalization=data.get("market_capitalization"),
            issued_shares=data.get("issued_shares"),
            is_delisting=data.get("is_delisting", False),
        )

    def to_dict(self) -> dict:
        return self.model_dump()

    def __eq__(self, other):
        if not isinstance(other, StockBrand):
            return False
        return self.code == other.code

    def __hash__(self):
        return hash(self.code)

    def __add__(self, other: "StockBrand") -> "StockBrand":
        if other is None:
            return self
        if type(other) is not StockBrand:
            raise KabutobashiEntityError()
        if self.code != other.code:
            raise KabutobashiEntityError()
        return StockBrand(
            code=self.code if self.code is not None else other.code,
            unit=self.unit if self.unit is not None else other.unit,
            market=self.market if self.market is not None else other.market,
            industry_type=self.industry_type if self.industry_type is not None else other.industry_type,
            market_capitalization=self.market_capitalization
            if self.market_capitalization is not None
            else other.market_capitalization,
            name=self.name if self.name is not None else other.name,
            issued_shares=self.issued_shares if self.issued_shares is not None else other.issued_shares,
            is_delisting=self.is_delisting or other.is_delisting,
        )

    # TODO modify
    # class Config:
    #     orm_mode = True


class StockPriceRecord(BaseModel, IDictSerialize, IDfSerialize):
    """
    Model: Entity
    JP: 日次株価
    """

    id: int = Field(description="id")
    code: str = Field(description="銘柄コード")
    dt: str = Field(description="日付")
    open: float = Field(description="始値")
    high: float = Field(description="高値")
    low: float = Field(description="底値")
    close: float = Field(description="終値")
    volume: int = Field(description="出来高")

    def __init__(
        self,
        id: Optional[int],
        code: str,
        dt: str,
        open: float,
        high: float,
        low: float,
        close: float,
        volume: int,
    ):

        super().__init__(
            id=0 if id is None else id,
            code=code,
            open=convert_float(open),
            high=convert_float(high),
            low=convert_float(low),
            close=convert_float(close),
            volume=convert_float(volume),
            dt=dt,
        )

    def is_outlier(self) -> bool:
        return self.open == 0 or self.high == 0 or self.low == 0 or self.close == 0

    def to_dict(self) -> dict:
        data = self.model_dump(exclude={"id"})
        return data

    @staticmethod
    def from_dict(data: dict) -> "StockPriceRecord":
        # code may "100.0"
        code = str(data["code"]).split(".")[0]
        return StockPriceRecord(
            id=data.get("id"),
            code=code,
            open=data["open"],
            high=data["high"],
            low=data["low"],
            close=data["close"],
            volume=data["volume"],
            dt=data["dt"],
        )

    def to_df(self) -> pd.DataFrame:
        return pd.DataFrame([self.to_dict()])

    @staticmethod
    def from_df(data: pd.DataFrame) -> List["StockPriceRecord"]:
        required_cols = ["open", "high", "low", "close", "code", "dt", "volume"]
        if set(required_cols) - set(data.columns):
            raise ValueError()
        if len(set(data["code"])) != 1:
            raise ValueError()

        records = []
        for _, row in data.iterrows():
            record = StockPriceRecord.from_dict(row)
            if record.is_valid_date():
                records.append(record)
        return records

    def is_valid_date(self) -> bool:
        return not jpholiday.is_holiday(datetime.strptime(self.dt, "%Y-%m-%d"))

    def __eq__(self, other):
        if not isinstance(other, StockPriceRecord):
            return False
        return self.code == other.code and self.dt == other.dt

    def __hash__(self):
        return hash(self.code)

    # TODO update
    # class Config:
    #     orm_mode = True


class StockReferenceIndicator(BaseModel, IDictSerialize):
    """
    Model: Entity
    JP: 参考指標
    """

    id: Optional[int]
    code: str = Field(description="銘柄コード")
    dt: str = Field(description="日付")
    psr: Optional[float] = Field(description="株価売上高倍率:Price to Sales Ratio", default=None)
    per: Optional[float] = Field(description="株価収益率:Price Earnings Ratio", default=None)
    pbr: Optional[float] = Field(description="株価純資産倍率:Price Book-value Ratio", default=None)
    ipo_manager: Optional[str] = Field(description="IPO_主幹", default=None)
    ipo_evaluation: Optional[str] = Field(description="IPO_評価", default=None)
    stock_listing_at: Optional[str] = Field(description="上場日", default=None)
    initial_price: Optional[float] = Field(description="初値", default=None)

    def __init__(self, id: int, code: str, dt: str, psr: Optional[float], per: Optional[float], pbr: Optional[float]):
        super().__init__(
            id=id,
            code=code,
            dt=dt,
            psr=psr,
            per=per,
            pbr=pbr,
        )

    def to_dict(self) -> dict:
        return self.model_dump()

    @staticmethod
    def from_dict(data: dict) -> "StockReferenceIndicator":
        return StockReferenceIndicator(
            id=0,
            code=data["code"],
            dt=data["dt"],
            psr=convert_float(data.get("psr")),
            per=convert_float(data.get("per")),
            pbr=convert_float(data.get("pbr")),
        )

    def __add__(self, other: "StockReferenceIndicator") -> "StockReferenceIndicator":
        if other is None:
            return self
        if type(other) is not StockReferenceIndicator:
            raise KabutobashiEntityError()
        if self.code != other.code:
            raise KabutobashiEntityError()
        return StockReferenceIndicator(
            id=self.id,
            code=self.code,
            dt=self.dt,
            psr=self.psr if self.psr is not None else other.psr,
            pbr=self.pbr if self.pbr is not None else other.pbr,
            per=self.per if self.per is not None else other.per,
        )


class Stock(BaseModel, IDfSerialize):
    """
    Model: Entity
    JP: 株
    """

    code: str = Field(description="銘柄コード")
    brand: StockBrand = Field(description="銘柄情報")
    daily_price_records: List[StockPriceRecord] = Field(description="日次株価記録", repr=False)
    reference_indicator: Optional[StockReferenceIndicator] = Field(description="参考指標")
    start_at: Optional[str] = Field(description="収集開始日")
    end_at: Optional[str] = Field(description="最新日時")

    def __init__(
        self,
        code: str,
        brand: StockBrand,
        daily_price_records: List[StockPriceRecord],
        reference_indicator: Optional[StockReferenceIndicator],
    ):
        if code != brand.code:
            raise KabutobashiEntityError()
        if reference_indicator is not None:
            if code != reference_indicator.code:
                raise KabutobashiEntityError()
        records_code_list = list(set([v.code for v in daily_price_records]))
        if len(records_code_list) > 1:
            raise KabutobashiEntityError()
        if records_code_list:
            if code != records_code_list[0]:
                raise KabutobashiEntityError()

        dt_list = [v.dt for v in daily_price_records]
        super().__init__(
            code=code,
            brand=brand,
            daily_price_records=daily_price_records,
            reference_indicator=reference_indicator,
            start_at=min(dt_list) if dt_list else None,
            end_at=max(dt_list) if dt_list else None,
        )

    def to_df(self, add_brand=False) -> pd.DataFrame:
        record_df = pd.concat([r.to_df() for r in self.daily_price_records])
        if add_brand:
            # from brand
            if self.brand.industry_type:
                record_df["industry_type"] = self.brand.industry_type
            if self.brand.market:
                record_df["market"] = self.brand.market
            if self.brand.market_capitalization:
                record_df["market_capitalization"] = self.brand.market_capitalization
            if self.brand.name:
                record_df["name"] = self.brand.name
            if self.brand.industry_type:
                record_df["industry_type"] = self.brand.industry_type
            if self.brand.issued_shares:
                record_df["issued_shares"] = self.brand.issued_shares
            if self.brand.unit:
                record_df["unit"] = self.brand.unit
            record_df["is_delisting"] = self.brand.is_delisting
        # from reference-indicator
        if self.reference_indicator.pbr:
            record_df["pbr"] = self.reference_indicator.pbr
        if self.reference_indicator.per:
            record_df["per"] = self.reference_indicator.per
        if self.reference_indicator.psr:
            record_df["psr"] = self.reference_indicator.psr
        return record_df.convert_dtypes().reset_index(drop=True)

    @staticmethod
    def from_df(data: pd.DataFrame) -> "Stock":
        data = data.reset_index(drop=True)
        required_cols = ["open", "high", "low", "close", "code", "dt", "volume"]
        if set(required_cols) - set(data.columns):
            raise KabutobashiEntityError()
        if len(set(data["code"])) != 1:
            raise KabutobashiEntityError()

        code = str(data["code"][0])
        daily_price_records = StockPriceRecord.from_df(data=data)
        latest_dt = max(data["dt"])
        latest_info = data[data["dt"] == latest_dt].to_dict(orient="records")[0]
        latest_info.update({"code": code})
        return Stock(
            code=code,
            brand=StockBrand.from_dict(data=latest_info),
            daily_price_records=daily_price_records,
            reference_indicator=StockReferenceIndicator.from_dict(data=latest_info),
        )

    def contains_outlier(self) -> bool:
        return any([v.is_outlier() for v in self.daily_price_records])

    @staticmethod
    def reduce(stocks: List["Stock"]) -> "Stock":
        code_list = list(set([v.code for v in stocks]))
        if len(code_list) > 1:
            raise KabutobashiEntityError()
        merge_target = stocks[0]
        for v in stocks[1:]:
            merge_target = merge_target + v
        return merge_target

    def __add__(self, other: "Stock") -> "Stock":
        if other is None:
            return self
        if type(other) is not Stock:
            raise KabutobashiEntityError()
        if self.code != other.code:
            raise KabutobashiEntityError()
        daily_price_records = []
        if self.daily_price_records:
            daily_price_records.extend(self.daily_price_records)
        if other.daily_price_records:
            daily_price_records.extend(other.daily_price_records)

        return Stock(
            code=self.code,
            brand=self.brand + other.brand,
            daily_price_records=list(set(daily_price_records)),
            reference_indicator=self.reference_indicator + other.reference_indicator,
        )



from typing import List, NoReturn, Union

import requests  # type: ignore

from kabutobashi.domain.errors import KabutobashiPageError
from kabutobashi.domain.values import (
    IHtmlPageRepository,
    RawHtmlPage,
    RawHtmlPageStockInfo,
    RawHtmlPageStockInfoMultipleDaysMain,
    RawHtmlPageStockInfoMultipleDaysSub,
    RawHtmlPageStockIpo,
    UserAgent,
)

__all__ = [
    "HtmlPageBasicRepository",
    "StockInfoHtmlPageRepository",
    "StockIpoHtmlPageRepository",
    "StockInfoMultipleDaysHtmlPageRepository",
]


class HtmlPageBasicRepository(IHtmlPageRepository):
    """
    Model: Repository(Interface)
    """

    def __init__(self, urls: List[str], page_type: str):
        self.urls = urls
        self.page_type = page_type

    @staticmethod
    def from_url(url: str, page_type: str) -> "RawHtmlPage":
        user_agent = UserAgent.get_user_agent_header()
        r = requests.get(url, headers=user_agent)

        if r.status_code != 200:
            raise KabutobashiPageError(url=url)

        # 日本語に対応
        r.encoding = r.apparent_encoding
        return RawHtmlPage(html=r.text, page_type=page_type, url=url)

    def _html_page_read(self) -> List[RawHtmlPage]:
        return [self.from_url(url=url, page_type=self.page_type) for url in self.urls]

    def _read_hook(self, html_page: RawHtmlPage) -> RawHtmlPage:
        return html_page

    def _html_page_write(self, data: RawHtmlPage) -> NoReturn:
        pass


class StockInfoHtmlPageRepository(HtmlPageBasicRepository):
    """
    Model: Repository(Implemented)
    """

    def __init__(self, code: Union[int, str]):
        super().__init__(page_type="info", urls=[f"https://minkabu.jp/stock/{code}"])
        self.code = code

    def _read_hook(self, html_page_list: List[RawHtmlPage]) -> RawHtmlPageStockInfo:
        return RawHtmlPageStockInfo(
            code=self.code, html=html_page_list[0].html, page_type=self.page_type, url=self.urls[0]
        )


class StockIpoHtmlPageRepository(HtmlPageBasicRepository):
    """
    Model: Repository(Implemented)
    """

    def __init__(self, year: str):
        super().__init__(page_type="ipo", urls=[f"https://96ut.com/ipo/list.php?year={year}"])
        self.year = year

    def _read_hook(self, html_page_list: List[RawHtmlPage]) -> RawHtmlPageStockIpo:
        return RawHtmlPageStockIpo(
            html=html_page_list[0].html, page_type=self.page_type, url=self.urls[0], year=self.year
        )


class StockInfoMultipleDaysHtmlPageRepository(HtmlPageBasicRepository):
    """
    Model: Repository(Implemented)
    """

    def __init__(self, code: Union[int, str]):
        main_html = f"https://minkabu.jp/stock/{code}/daily_bar"
        sub_html = f"https://minkabu.jp/stock/{code}/daily_valuation"
        super().__init__(page_type="info_multiple", urls=[main_html, sub_html])
        self.code = code

    def _read_hook(
        self, html_page_list: List[RawHtmlPage]
    ) -> List[Union[RawHtmlPageStockInfoMultipleDaysMain, RawHtmlPageStockInfoMultipleDaysSub]]:
        return [
            RawHtmlPageStockInfoMultipleDaysMain(
                html=html_page_list[0].html, page_type=self.page_type, url=self.urls[0], code=self.code
            ),
            RawHtmlPageStockInfoMultipleDaysSub(
                html=html_page_list[1].html, page_type=self.page_type, url=self.urls[1], code=self.code
            ),
        ]



import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


class PctChangeProcess(ProcessMethod):
    """
    変化率を計算する
    """

    method_name: str = "pct_change"
    method_type: MethodType = MethodType.PARAMETERIZE

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["diff"] = -1
        # 正負が交差した点
        df_ = df_.join(self._cross(df_["diff"]))
        df_ = df_.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df_

    def _processed_columns(self) -> list:
        return []

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        pct_05 = df_x["close"].pct_change(5).mean()
        pct_10 = df_x["close"].pct_change(10).mean()
        pct_20 = df_x["close"].pct_change(20).mean()
        pct_30 = df_x["close"].pct_change(30).mean()
        pct_40 = df_x["close"].pct_change(40).mean()
        return {"pct_05": pct_05, "pct_10": pct_10, "pct_20": pct_20, "pct_30": pct_30, "pct_40": pct_40}


class PctChangeVisualize(VisualizeMethod):
    """
    変化率を計算する
    """

    def _color_mapping(self) -> list:
        return []

    def _visualize_option(self) -> dict:
        return {"position": "-"}


pct_change = Method.of(process_method=PctChangeProcess(), visualize_method=PctChangeVisualize())



import re
from dataclasses import dataclass
from datetime import datetime
from logging import getLogger
from typing import Dict, List, Union

import pandas as pd

from kabutobashi.domain.values import (
    DecodeHtmlPageStockInfoMinkabuTop,
    DecodeHtmlPageStockInfoMultipleDays,
    DecodeHtmlPageStockIpo,
    RawHtmlPageStockInfo,
    RawHtmlPageStockInfoMultipleDaysMain,
    RawHtmlPageStockInfoMultipleDaysSub,
    RawHtmlPageStockIpo,
)

from .utils import IHtmlDecoder, PageDecoder

logger = getLogger(__name__)


@dataclass(frozen=True)
class StockInfoMinkabuTopHtmlDecoder(IHtmlDecoder):
    """
    Model: Service(Implemented)

    Examples:
        >>> from kabutobashi.infrastructure.repository import StockInfoHtmlPageRepository
        >>> # get single page
        >>> html_page = StockInfoHtmlPageRepository(code="0001").read()
        >>> result = StockInfoMinkabuTopHtmlDecoder().decode_to_dict(html_page=html_page)
    """

    def _decode_to_object_hook(self, data: dict) -> DecodeHtmlPageStockInfoMinkabuTop:
        return DecodeHtmlPageStockInfoMinkabuTop.from_dict(data=data)

    def _decode(self, html_page: RawHtmlPageStockInfo) -> dict:
        soup = html_page.get_as_soup()
        result: Dict[str, Union[str, bool, int, float, List[str]]] = {"html": html_page.html}

        stock_board_tag = "md_stockBoard"

        raw_dt = PageDecoder(tag1="span", class1="fsm").decode(bs=soup)
        pattern = r"\((?P<month>[0-9]+)/(?P<day>[0-9]+)\)|\((?P<hour>[0-9]+):(?P<minute>[0-9]+)\)"
        match_result = re.match(pattern, raw_dt)
        dt = datetime.now()
        if match_result:
            rep = match_result.groupdict()
            if rep.get("month"):
                dt = dt.replace(month=int(rep["month"]))
            if rep.get("day"):
                dt = dt.replace(day=int(rep["day"]))

        # ページ上部の情報を取得
        stock_board = soup.find("div", {"class": stock_board_tag})
        result.update(
            {
                "stock_label": PageDecoder(tag1="div", class1="stock_label").decode(bs=stock_board),
                "name": PageDecoder(tag1="p", class1="md_stockBoard_stockName").decode(bs=stock_board),
                "close": PageDecoder(tag1="div", class1="stock_price").decode(bs=stock_board),
            }
        )

        # ページ中央の情報を取得
        stock_detail = soup.find("div", {"id": "main"})
        info = {}
        for li in stock_detail.find_all("tr", {"class": "ly_vamd"}):
            info[li.find("th").get_text()] = li.find("td").get_text()
        stock_label = str(result.get("stock_label", ""))
        code, market = stock_label.split("  ")
        result.update(
            {
                "dt": dt.strftime("%Y-%m-%d"),
                "code": code,
                "industry_type": PageDecoder(tag1="div", class1="ly_content_wrapper size_ss").decode(bs=stock_detail),
                "market": market,
                "open": info.get("始値", "0"),
                "high": info.get("高値", "0"),
                "low": info.get("安値", "0"),
                "unit": info.get("単元株数", "0"),
                "per": info.get("PER(調整後)", "0"),
                "psr": info.get("PSR", "0"),
                "pbr": info.get("PBR", "0"),
                "volume": info.get("出来高", "0"),
                "market_capitalization": info.get("時価総額", "---"),
                "issued_shares": info.get("発行済株数", "---"),
            }
        )

        return result


@dataclass(frozen=True)
class StockIpoHtmlDecoder(IHtmlDecoder):
    """
    Model: Service(Implemented)
    """

    def _decode(self, html_page: RawHtmlPageStockIpo) -> dict:
        soup = html_page.get_as_soup()
        table_content = soup.find("div", {"class": "tablewrap"})
        table_thead = table_content.find("thead")
        # headの取得
        table_head_list = []
        for th in table_thead.find_all("th"):
            table_head_list.append(th.get_text())

        # bodyの取得
        table_tbody = table_content.find("tbody")
        whole_result = []
        for idx, tr in enumerate(table_tbody.find_all("tr")):
            table_body_dict = {}
            for header, td in zip(table_head_list, tr.find_all("td")):
                table_body_dict[header] = td.get_text().replace("\n", "")
            whole_result.append(DecodeHtmlPageStockIpo.from_dict(data=table_body_dict).to_dict())
        return {"ipo_list": whole_result}

    def _decode_to_object_hook(self, data: dict) -> List[DecodeHtmlPageStockIpo]:
        ipo_list = data["ipo_list"]
        result_list = []
        for v in ipo_list:
            result_list.append(DecodeHtmlPageStockIpo.from_dict(data=v))
        return result_list


@dataclass(frozen=True)
class StockInfoMultipleDaysHtmlDecoder(IHtmlDecoder):
    """
    Model: Service(Implemented)

    Examples:
        >>> from kabutobashi.infrastructure.repository import StockInfoMultipleDaysHtmlPageRepository
        >>> from kabutobashi.domain.services import StockInfoMultipleDaysHtmlDecoder
        >>> import kabutobashi as kb
        >>> html_page_list = StockInfoMultipleDaysHtmlPageRepository(code=1375).read()
        >>> data = StockInfoMultipleDaysHtmlDecoder().decode_to_dict(html_page=html_page_list)
        >>> df = pd.DataFrame(data)
        >>> records = kb.Stock.from_df(df)
    """

    def _decode(
        self, html_page: List[Union[RawHtmlPageStockInfoMultipleDaysMain, RawHtmlPageStockInfoMultipleDaysSub]]
    ) -> dict:
        result_1 = []
        result_2 = []
        main_soup = html_page[0].get_as_soup()
        sub_soup = html_page[1].get_as_soup()
        stock_recordset_tag = "md_card md_box"

        # ページの情報を取得
        stock_recordset = main_soup.find("div", {"class": stock_recordset_tag})
        mapping = {0: "dt", 1: "open", 2: "high", 3: "low", 4: "close", 5: "調整後終値", 6: "volume"}
        for tr in stock_recordset.find_all("tr"):
            tmp = {}
            for idx, td in enumerate(tr.find_all("td")):
                tmp.update({mapping[idx]: td.get_text()})
            result_1.append(tmp)

        # そのほかの情報
        stock_board = sub_soup.find("div", {"class": "md_card md_box mzp"})
        mapping2 = {0: "dt", 1: "psr", 2: "per", 3: "pbr", 4: "配当利回り(%)", 5: "close", 6: "調整後終値", 7: "volume"}
        for tr in stock_board.find_all("tr"):
            tmp = {}
            for idx, td in enumerate(tr.find_all("td")):
                tmp.update({mapping2[idx]: td.get_text()})
            result_2.append(tmp)

        df1 = pd.DataFrame(result_1).dropna()
        df2 = pd.DataFrame(result_2).dropna()

        df1 = df1[["dt", "open", "high", "low", "close"]]
        df2 = df2[["dt", "psr", "per", "pbr", "volume"]]

        df = pd.merge(df1, df2, on="dt")
        df["code"] = html_page[0].code
        return {"info_list": df.to_dict(orient="records")}

    def _decode_to_object_hook(self, data: dict) -> List["DecodeHtmlPageStockInfoMultipleDays"]:
        info_list = data["info_list"]
        result_list = []
        for v in info_list:
            result_list.append(DecodeHtmlPageStockInfoMultipleDays.from_dict(data=v))
        return result_list



# pro = stock_agg.with_estimated(ANALYSIS)

import pytest

import kabutobashi as kb


@pytest.fixture(scope="module", autouse=True)
def stock_agg() -> kb.StockCodeSingleAggregate:
    df = kb.example()
    yield kb.StockCodeSingleAggregate.of(entity=df, code="1375")


def test_analysis_with_fundamental(stock_agg):
    with pytest.raises(KeyError):
        _ = stock_agg.with_processed([]).with_estimated(kb.stock_analysis)
    with pytest.raises(KeyError):
        _ = stock_agg.with_processed([kb.sma]).with_estimated(kb.stock_analysis)
    with pytest.raises(KeyError):
        _ = stock_agg.with_processed([kb.sma, kb.macd]).with_estimated(kb.stock_analysis)
    with pytest.raises(KeyError):
        _ = stock_agg.with_processed([kb.sma, kb.macd, kb.stochastics]).with_estimated(kb.stock_analysis)
    with pytest.raises(KeyError):
        _ = stock_agg.with_processed([kb.sma, kb.macd, kb.stochastics, kb.bollinger_bands]).with_estimated(
            kb.stock_analysis
        )
    with pytest.raises(KeyError):
        _ = stock_agg.with_processed([kb.sma, kb.macd, kb.stochastics, kb.bollinger_bands, kb.momentum]).with_estimated(
            kb.stock_analysis
        )
    with pytest.raises(KeyError):
        _ = stock_agg.with_processed(
            [kb.sma, kb.macd, kb.stochastics, kb.bollinger_bands, kb.momentum, kb.psycho_logical]
        ).with_estimated(kb.stock_analysis)



import pandas as pd

from .method import Method, MethodType, ProcessMethod


class BasicProcess(ProcessMethod):
    """
    株のvolumeやPBR, PSR, PERなどの値を返す。
    parameterizeのみに利用される。
    """

    method_name: str = "basic"
    method_type: MethodType = MethodType.PARAMETERIZE

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["diff"] = -1
        # 正負が交差した点
        df_ = df_.join(self._cross(df_["diff"]))
        df_ = df_.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df_

    def _processed_columns(self) -> list:
        return []

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        try:
            pbr = float(list(df_x["pbr"])[-1])
        except ValueError:
            pbr = 0
        except TypeError:
            pbr = 0
        try:
            per = float(list(df_x["per"])[-1])
        except ValueError:
            per = 0
        except TypeError:
            per = 0
        try:
            psr = float(list(df_x["psr"])[-1])
        except ValueError:
            psr = 0
        except TypeError:
            psr = 0
        try:
            volume = float(list(df_x["volume"])[-1])
        except ValueError:
            volume = 0
        except TypeError:
            volume = 0
        return {"pbr": pbr, "psr": psr, "per": per, "volume": volume}


basic = Method.of(process_method=BasicProcess(), visualize_method=None)



from datetime import datetime, timedelta
from typing import Union

import jpholiday

from kabutobashi.domain.errors import KabutobashiBaseError, KabutobashiEntityError


def get_past_n_days(current_date: str, n: int = 60) -> list:
    """
    土日と祝日を考慮したn営業日前までの日付のリストを返す関数

    Args:
        current_date: n日前を計算する起点となる日
        n: n日前

    Returns:
        date list, ex ["%Y-%m-%d", "%Y-%m-%d", "%Y-%m-%d", ...]
    """
    multiply_list = [2, 4, 8, 16]
    for multiply in multiply_list:
        return_candidate = _get_past_n_days(current_date=current_date, n=n, multiply=multiply)
        if len(return_candidate) == n:
            return return_candidate
    raise KabutobashiBaseError(f"{n}日前を正しく取得できませんでした")


def _get_past_n_days(current_date: str, n: int, multiply: int) -> list:
    """
    n*multiplyの日数分のうち、商取引が行われる日を取得する

    Args:
        current_date: n日前を計算する起点となる日
        n: n日前
        multiply: n日前にかける数。
    """
    end_date = datetime.strptime(current_date, "%Y-%m-%d")
    # 2倍しているのは土日や祝日が排除されるため
    # また、nが小さすぎると休日が重なった場合に日数の取得ができないため
    back_n_days = n * multiply
    date_candidate = [end_date - timedelta(days=d) for d in range(back_n_days)]
    # 土日を除く
    filter_weekend = [d for d in date_candidate if d.weekday() < 5]
    # 祝日を除く
    filter_holiday = [d for d in filter_weekend if not jpholiday.is_holiday(d)]
    # 文字列に日付を変えてreturn
    return list(map(lambda x: x.strftime("%Y-%m-%d"), filter_holiday[:n]))


def replace(input_value: str) -> str:
    if input_value == "-":
        return "0"
    elif input_value == "":
        return "0"
    return input_value.replace("---", "0").replace("円", "").replace("株", "").replace("倍", "").replace(",", "")


def convert_float(input_value: Union[str, float, int]) -> float:
    if type(input_value) is float:
        return input_value
    elif type(input_value) is int:
        return float(input_value)
    elif type(input_value) is str:
        try:
            return float(replace(input_value=input_value))
        except ValueError as e:
            raise KabutobashiEntityError(f"cannot convert {input_value} to float: {e}")
    raise KabutobashiEntityError(f"cannot convert {input_value} to float")


def convert_int(input_value: Union[str, float, int]) -> int:
    if type(input_value) == int:
        return input_value
    elif type(input_value) == float:
        try:
            return int(input_value)
        except ValueError:
            return 0
    elif type(input_value) is str:
        try:
            return int(replace(input_value=input_value))
        except ValueError as e:
            raise KabutobashiEntityError(f"cannot convert {input_value} to integer: {e}")
    else:
        try:
            return int(input_value)
        except Exception:
            raise KabutobashiEntityError(f"cannot convert {input_value} to int")



from dataclasses import dataclass, field
from typing import List, Optional, Union

import pandas as pd

from kabutobashi.domain.entity import Stock
from kabutobashi.domain.errors import KabutobashiEntityError
from kabutobashi.domain.serialize import IDictSerialize
from kabutobashi.domain.services.analyze import StockAnalysis
from kabutobashi.domain.services.method import Method
from kabutobashi.domain.values import StockDataEstimated, StockDataProcessed, StockDataVisualized

__all__ = ["StockCodeSingleAggregate"]


@dataclass(frozen=True)
class StockCodeSingleAggregate(IDictSerialize):
    """
    StockCodeSingleAggregate: Aggregate

    Examples:
        >>> import kabutobashi as kb
        >>> import pandas as pd
        >>> data_list = []
        >>> df = kb.example()
        >>> methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        >>> stock_analysis = kb.stock_analysis
        >>>
        >>> agg = kb.StockCodeSingleAggregate.of(entity=df, code="1234")
        >>>     .with_processed(methods)
        >>>     .with_estimated(stock_analysis)
        >>> print(agg.weighted_estimated_value())
    """

    code: str
    stock: Stock
    processed_list: List[StockDataProcessed] = field(default_factory=list, repr=False)
    estimated_list: List[StockDataEstimated] = field(default_factory=list, repr=False)

    def __post_init__(self):
        pass

    @staticmethod
    def of(entity: Union[pd.DataFrame, Stock], *, code: Optional[str] = None) -> "StockCodeSingleAggregate":
        if type(entity) is pd.DataFrame:
            entity["code"] = entity["code"].astype(str)
            stock = Stock.from_df(data=entity[entity["code"] == code])
        elif type(entity) is Stock:
            stock = entity
        else:
            raise KabutobashiEntityError("accept pd.DataFrame or StockDataSingleCode")

        code = stock.code
        return StockCodeSingleAggregate(code=code, stock=stock)

    def with_processed(self, methods: List[Method]) -> "StockCodeSingleAggregate":
        # check all methods
        for method in methods:
            if not isinstance(method, Method):
                raise KabutobashiEntityError()

        df = self.stock.to_df()
        return StockCodeSingleAggregate(
            code=self.code,
            stock=self.stock,
            processed_list=[m.process_method.process(df=df) for m in methods],
        )

    def _to_single_estimated(self, stock_analysis: StockAnalysis) -> StockDataEstimated:
        def get_parameters():
            data_ = {}
            for a in self.processed_list:
                data_.update(a.parameters)
            return data_

        return StockDataEstimated(
            code=self.code,
            estimate_filter_name=stock_analysis.estimate_filter_name,
            estimated_value=stock_analysis.estimate(data=get_parameters()),
        )

    def with_estimated(self, stock_analysis: List[StockAnalysis]) -> "StockCodeSingleAggregate":
        """

        Returns:
            StockDataEstimatedByMultipleFilter
        """
        return StockCodeSingleAggregate(
            code=self.code,
            stock=self.stock,
            processed_list=self.processed_list,
            estimated_list=[self._to_single_estimated(stock_analysis=sa) for sa in stock_analysis],
        )

    def weighted_estimated_value(self, weights: dict = None) -> float:
        if not weights:
            weights = {}
        return sum([e.weighted_estimated_value(weights=weights) for e in self.estimated_list])

    def estimate_filter_concat_name(self) -> str:
        estimate_filter_names = sorted([e.estimate_filter_name for e in self.estimated_list])
        return "_".join(estimate_filter_names)

    def visualize(self, method: Method, size_ratio: int = 2) -> StockDataVisualized:
        df = self.stock.to_df()
        processed_df = method.process_method.process(df=df)
        if method.visualize_method:
            return method.visualize_method.visualize(size_ratio=size_ratio, df=processed_df.df)
        else:
            raise KabutobashiEntityError(f"method {method.process_method.method_name} has no visualize method")

    def to_dict(self) -> dict:
        return {
            "code": self.code,
            "name": self.stock.brand.name,
            "market": self.stock.brand.market,
            "industry_type": self.stock.brand.industry_type,
            "weighted_estimated_value": self.weighted_estimated_value(),
            "estimate_name": self.estimate_filter_concat_name(),
        }

    @staticmethod
    def from_dict(data: dict):
        raise NotImplementedError("The method would not be implemented.")



from typing import List

import pandas as pd

from kabutobashi.domain.entity.stock import Stock, StockBrand


def decode_brand_list(path: str) -> List[Stock]:
    """
    See Also: https://www.jpx.co.jp/markets/statistics-equities/misc/01.html
    """
    df = pd.read_excel(path)
    column_renames = {"日付": "dt", "コード": "code", "銘柄名": "name", "市場・商品区分": "market", "33業種区分": "industry_type"}
    df = df.rename(columns=column_renames)
    df = df[column_renames.values()]
    df["market"] = df["market"].apply(lambda x: x.replace("（内国株式）", ""))
    prime_df = df[df["market"] == "プライム"]
    standard_df = df[df["market"] == "スタンダード"]
    growth_df = df[df["market"] == "グロース"]
    merged_df = pd.concat([prime_df, standard_df, growth_df]).reset_index()

    stock_list = []
    for idx, row in merged_df.iterrows():
        brand = StockBrand.from_dict(dict(row))
        stock = Stock(code=brand.code, brand=brand, daily_price_records=[], reference_indicator=None)
        stock_list.append(stock)

    return stock_list



from typing import List, Union

from kabutobashi.domain.entity.stock import Stock, StockBrand, StockPriceRecord, StockReferenceIndicator
from kabutobashi.domain.values import (
    DecodeHtmlPageStockInfoMinkabuTop,
    DecodeHtmlPageStockInfoMultipleDays,
    DecodeHtmlPageStockIpo,
)


class StockConverter:
    def convert(
        self,
        value_object: Union[
            DecodeHtmlPageStockIpo, DecodeHtmlPageStockInfoMinkabuTop, List[DecodeHtmlPageStockInfoMultipleDays]
        ],
    ) -> Stock:
        if type(value_object) is DecodeHtmlPageStockInfoMinkabuTop:
            return self._convert_stock_info_minkabu_top_page(value_object=value_object)
        elif type(value_object) is DecodeHtmlPageStockIpo:
            return self._convert_stock_ipo_page(value_object=value_object)
        elif type(value_object) is list:
            if all([type(v) is DecodeHtmlPageStockInfoMultipleDays for v in value_object]):
                return self._convert_stock_info_multiple_days_list(value_object=value_object)
        else:
            raise ValueError()

    @staticmethod
    def _convert_stock_info_minkabu_top_page(value_object: DecodeHtmlPageStockInfoMinkabuTop) -> Stock:
        data = value_object.to_dict()

        stock_brand = StockBrand.from_dict(data)
        stock_reference_indicator = StockReferenceIndicator.from_dict(data)

        return Stock(
            code=value_object.code,
            brand=stock_brand,
            daily_price_records=[StockPriceRecord.from_dict(data)],
            reference_indicator=stock_reference_indicator,
        )

    @staticmethod
    def _convert_stock_info_multiple_days(value_object: DecodeHtmlPageStockInfoMultipleDays) -> Stock:
        data = value_object.to_dict()

        stock_brand = StockBrand.from_dict(data)
        stock_reference_indicator = StockReferenceIndicator.from_dict(data)
        return Stock(
            code=value_object.code,
            brand=stock_brand,
            daily_price_records=[StockPriceRecord.from_dict(data)],
            reference_indicator=stock_reference_indicator,
        )

    @staticmethod
    def _convert_stock_info_multiple_days_list(value_object: List[DecodeHtmlPageStockInfoMultipleDays]) -> Stock:
        return Stock.reduce(stocks=[StockConverter._convert_stock_info_multiple_days(v) for v in value_object])

    @staticmethod
    def _convert_stock_ipo_page(value_object: DecodeHtmlPageStockIpo) -> Stock:

        data = value_object.to_dict()
        stock_brand = StockBrand.from_dict(data)

        return Stock(
            code=value_object.code,
            brand=stock_brand,
            daily_price_records=[],
            reference_indicator=None,
        )



from typing import Optional

from injector import Binder, Module


class StockCrawlDi(Module):
    """
    Model: DI Container (for crawl and decode)
    """

    def __init__(self, page_type: str, *, code: Optional[str] = None, year: Optional[str] = None):
        self.page_type = page_type
        self.code = code
        self.year = year

    def configure(self, binder: Binder) -> None:
        # repository/decoders
        from kabutobashi.domain.services import (
            IHtmlDecoder,
            StockInfoMinkabuTopHtmlDecoder,
            StockInfoMultipleDaysHtmlDecoder,
            StockIpoHtmlDecoder,
        )
        from kabutobashi.domain.values import IHtmlPageRepository
        from kabutobashi.infrastructure.repository import (
            StockInfoHtmlPageRepository,
            StockInfoMultipleDaysHtmlPageRepository,
            StockIpoHtmlPageRepository,
        )

        if self.page_type == "info":
            # repository-bind
            binder.bind(IHtmlPageRepository, StockInfoHtmlPageRepository(code=self.code))
            # decoder-bind
            binder.bind(IHtmlDecoder, StockInfoMinkabuTopHtmlDecoder)
        elif self.page_type == "ipo":
            # repository-bind
            binder.bind(IHtmlPageRepository, StockIpoHtmlPageRepository(year=self.year))
            # decoder-bind
            binder.bind(IHtmlDecoder, StockIpoHtmlDecoder)
        elif self.page_type == "info_multiple":
            # repository-bind
            binder.bind(IHtmlPageRepository, StockInfoMultipleDaysHtmlPageRepository(code=self.code))
            # decoder-bind
            binder.bind(IHtmlDecoder, StockInfoMultipleDaysHtmlDecoder)



import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


class MomentumProcess(ProcessMethod):
    """
    See Also:
        https://www.sevendata.co.jp/shihyou/technical/momentum.html
    """

    term: int = 25
    method_name: str = "momentum"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            momentum=df["close"].shift(10),
        ).fillna(0)
        df = df.assign(sma_momentum=lambda x: x["momentum"].rolling(self.term).mean())
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.join(self._cross(df["sma_momentum"]))
        df = df.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df

    def _processed_columns(self) -> list:
        return ["momentum", "sma_momentum"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return {}


class MomentumVisualize(VisualizeMethod):
    """
    See Also:
        https://www.sevendata.co.jp/shihyou/technical/momentum.html
    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "momentum", "color": "", "label": "momentum"},
            {"df_key": "sma_momentum", "color": "", "label": "sma_momentum"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "lower"}


momentum = Method.of(process_method=MomentumProcess(), visualize_method=MomentumVisualize())



from .analyze import SaFundamental, SaVolume, StockAnalysis
from .converter import StockConverter
from .decode_html import (
    IHtmlDecoder,
    StockInfoMinkabuTopHtmlDecoder,
    StockInfoMultipleDaysHtmlDecoder,
    StockIpoHtmlDecoder,
)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import reduce
from logging import getLogger
from typing import List, Optional, Union

from bs4 import BeautifulSoup

from kabutobashi.domain.values import RawHtmlPage

logger = getLogger(__name__)
__all__ = ["IHtmlDecoder", "PageDecoder"]


@dataclass(frozen=True)
class PageDecoder:
    tag1: Optional[str] = None
    class1: Optional[str] = None
    id1: Optional[str] = None
    default: str = ""

    def _decode(self, value):
        class1 = {"class": self.class1}

        set_value = None
        # tag1から取得
        if self.tag1 is not None:
            if class1["class"] is not None:
                set_value = value.find(self.tag1, self.class1)
            else:
                set_value = value.find(self.tag1)

        if set_value is None:
            return self.default

        # 文字列を置換して保持
        return self.replace(set_value.get_text())

    def decode(self, bs: BeautifulSoup) -> Union[str, List[str]]:
        return self._decode(value=bs)

    @staticmethod
    def replace(input_text: str) -> str:
        target_list = [" ", "\t", "\n", "\r", "円"]

        def remove_of(_input: str, target: str):
            return _input.replace(target, "")

        result = reduce(remove_of, target_list, input_text)
        return result.replace("\xa0", " ")


@dataclass(frozen=True)  # type: ignore
class IHtmlDecoder(ABC):
    """
    Model: Service(Interface)
    JP: HTML変換サービス
    """

    @abstractmethod
    def _decode(self, html_page: RawHtmlPage) -> dict:
        raise NotImplementedError()  # pragma: no cover

    def decode_to_dict(self, html_page: RawHtmlPage) -> dict:
        return self._decode(html_page=html_page)

    @abstractmethod
    def _decode_to_object_hook(self, data: dict) -> object:
        raise NotImplementedError()  # pragma: no cover

    def decode_to_object(self, html_page: RawHtmlPage) -> object:
        data = self._decode(html_page=html_page)
        return self._decode_to_object_hook(data=data)



from abc import ABC, abstractmethod
from typing import List, NoReturn, Union

from bs4 import BeautifulSoup
from pydantic import BaseModel, Field, field_validator


class RawHtmlPage(BaseModel):
    """
    Model: ValueObject
    JP: 変換対象HTML
    """

    html: str = Field(repr=False)
    page_type: str
    url: str

    @classmethod
    @field_validator("page_type")
    def _validate_page_type(cls, v):
        assert v in ["info", "info_multiple", "ipo"]

    def get_as_soup(self) -> BeautifulSoup:
        return BeautifulSoup(self.html, features="lxml")


class IHtmlPageRepository(ABC):
    """
    Model: Repository(for ValueObject, Interface)
    """

    def read(self) -> Union[RawHtmlPage, List[RawHtmlPage]]:
        html_page_list = self._html_page_read()
        return self._read_hook(html_page_list=html_page_list)

    def _read_hook(self, html_page_list: List[RawHtmlPage]) -> Union[RawHtmlPage, List[RawHtmlPage]]:
        return html_page_list

    @abstractmethod
    def _html_page_read(self) -> Union[RawHtmlPage, List[RawHtmlPage]]:
        raise NotImplementedError()  # pragma: no cover

    def write(self, data: RawHtmlPage) -> NoReturn:
        self._html_page_write(data=data)

    @abstractmethod
    def _html_page_write(self, data: RawHtmlPage) -> NoReturn:
        raise NotImplementedError()  # pragma: no cover


class RawHtmlPageStockInfo(RawHtmlPage):
    code: Union[int, str]


class RawHtmlPageStockIpo(RawHtmlPage):
    year: str


class RawHtmlPageStockInfoMultipleDaysMain(RawHtmlPage):
    code: Union[int, str]


class RawHtmlPageStockInfoMultipleDaysSub(RawHtmlPage):
    code: Union[int, str]



import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


class IchimokuProcess(ProcessMethod):
    """

    See Also:
        https://kabu.com/investment/guide/technical/04.html
    """

    short_term: int = 12
    medium_term: int = 26
    long_term: int = 52
    method_name: str = "ichimoku"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            # 短期の線
            short_max=lambda x: x["close"].rolling(self.short_term).max(),
            short_min=lambda x: x["close"].rolling(self.short_term).min(),
            # 中期の線
            medium_max=lambda x: x["close"].rolling(self.medium_term).max(),
            medium_min=lambda x: x["close"].rolling(self.medium_term).min(),
            # 長期線
            long_max=lambda x: x["close"].rolling(self.long_term).max(),
            long_min=lambda x: x["close"].rolling(self.long_term).min(),
        )

        # 指標の計算
        df = df.assign(
            line_change=lambda x: (x["short_max"] + x["short_min"]) / 2,
            line_base=lambda x: (x["medium_max"] + x["medium_min"]) / 2,
            # 先行線
            proceeding_span_1=lambda x: (x["line_change"] + x["line_base"]) / 2,
            proceeding_span_2=lambda x: (x["long_max"] + x["long_min"]) / 2,
        )

        # 値のshift
        df = df.assign(
            proceeding_span_1=df["proceeding_span_1"].shift(26),
            proceeding_span_2=df["proceeding_span_2"].shift(26),
            delayed_span=df["close"].shift(26),
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def _processed_columns(self) -> list:
        return ["line_change", "line_base", "proceeding_span_1", "proceeding_span_2", "delayed_span"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return {}


class IchimokuVisualize(VisualizeMethod):
    """

    See Also:
        https://kabu.com/investment/guide/technical/04.html
    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "", "color": "", "label": ""},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "in"}


ichimoku = Method.of(process_method=IchimokuProcess(), visualize_method=IchimokuVisualize())



import os
import sys

import pytest

# テスト対象のファイルへのパスを通している
# pytestの設定
PARENT_PATH = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
SOURCE_PATH = PARENT_PATH.rsplit("/", 1)[0]
print(f"test_target_path:= {SOURCE_PATH}")
sys.path.append(f"{SOURCE_PATH}")

# テスト対象のディレクトリに移動
os.chdir("./kabutobashi")

# skip if: https://thinkami.hatenablog.com/entry/2017/10/25/222551

DATA_PATH = f"{SOURCE_PATH}/data"


@pytest.fixture
def data_path():
    yield DATA_PATH



import setuptools
import os


def get_version(init_file_path):
    version_line = list(filter(lambda l: l.startswith("VERSION"), open(init_file_path)))[0]

    # eval is required to convert from string to tuple,
    # because VERSION defined in __init__.py is tuple
    version_tuple = eval(version_line.split("=")[-1])

    # join with dot
    return ".".join(map(str, version_tuple))


with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

# get __version__ from __init__.py
init = os.path.join(os.path.dirname(__file__), "kabutobashi", "__init__.py")
VERSION = get_version(init_file_path=init)


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kabutobashi",
    version=VERSION,
    author="gsy0911",
    author_email="yoshiki0911@gmail.com",
    description="Analyze stock",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gsy0911/kabutobashi",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.8",
)



from dataclasses import dataclass

from injector import Injector, inject

from kabutobashi.domain.services import IHtmlDecoder, StockConverter
from kabutobashi.domain.values import IHtmlPageRepository

from .di_container import StockCrawlDi


@inject
@dataclass
class DataCrawlController:
    html_page_repository: IHtmlPageRepository
    html_decoder: IHtmlDecoder

    def run(self):
        # code: str
        page_html = self.html_page_repository.read()
        return self.html_decoder.decode_to_object(html_page=page_html)


def crawl_info(code: str):
    di = Injector([StockCrawlDi(page_type="info", code=code)])
    data_crawler = di.get(DataCrawlController)
    info_object = data_crawler.run()
    return StockConverter().convert(value_object=info_object)


def crawl_info_multiple(code: str):
    di = Injector([StockCrawlDi(page_type="info_multiple", code=code)])
    data_crawler = di.get(DataCrawlController)
    info_object = data_crawler.run()
    return StockConverter().convert(value_object=info_object)


def crawl_ipo(year: str):
    di = Injector([StockCrawlDi(page_type="ipo", year=year)])
    data_crawler = di.get(DataCrawlController)
    ipo_list_object = data_crawler.run()
    return [StockConverter().convert(value_object=v) for v in ipo_list_object]



from abc import ABC

from pydantic import BaseModel, Field, computed_field

from kabutobashi.utilities import convert_float

__all__ = [
    "DecodedHtmlPage",
    "DecodeHtmlPageStockIpo",
    "DecodeHtmlPageStockInfoMultipleDays",
    "DecodeHtmlPageStockInfoMinkabuTop",
]


class DecodedHtmlPage(ABC):
    """
    Model: ValueObject
    JP: 変換済みHTML
    id: code & dt
    """


class DecodeHtmlPageStockInfoMinkabuTop(BaseModel, DecodedHtmlPage):
    code: str
    dt: str
    name: str
    industry_type: str
    market: str
    open: str
    high: str
    low: str
    close: str
    unit: str
    per: str
    psr: str
    pbr: str
    volume: str
    market_capitalization: str
    issued_shares: str
    html: str = Field(repr=False)

    @computed_field
    @property
    def is_delisting(self) -> int:
        if self.open == "---" and self.high == "---" and self.low == "---":
            return True
        return False

    def to_dict(self) -> dict:
        return self.model_dump(exclude={"html"})

    @staticmethod
    def from_dict(data: dict) -> "DecodeHtmlPageStockInfoMinkabuTop":
        return DecodeHtmlPageStockInfoMinkabuTop(
            code=data["code"],
            dt=data["dt"],
            name=data["name"],
            open=data["open"],
            high=data["high"],
            low=data["low"],
            close=data["close"],
            pbr=data["pbr"],
            per=data["per"],
            psr=data["psr"],
            unit=data["unit"],
            volume=data["volume"],
            market=data["market"],
            market_capitalization=data["market_capitalization"],
            industry_type=data["industry_type"],
            issued_shares=data["issued_shares"],
            html=data["html"],
        )


class DecodeHtmlPageStockInfoMultipleDays(BaseModel, DecodedHtmlPage):
    code: str
    dt: str
    open: str
    high: str
    low: str
    close: str
    per: str
    psr: str
    pbr: str
    volume: str
    html: str = Field(repr=False)

    def to_dict(self) -> dict:
        return self.model_dump(exclude={"html"})

    @staticmethod
    def from_dict(data: dict) -> "DecodeHtmlPageStockInfoMultipleDays":
        return DecodeHtmlPageStockInfoMultipleDays(
            code=data["code"],
            dt=data["dt"],
            open=data["open"],
            high=data["high"],
            low=data["low"],
            close=data["close"],
            pbr=data["pbr"],
            per=data["per"],
            psr=data["psr"],
            volume=data["volume"],
            html="",
        )


class DecodeHtmlPageStockIpo(BaseModel, DecodedHtmlPage):
    code: str = Field(description="銘柄コード")
    manager: str = Field(description="主幹")
    stock_listing_at: str = Field(description="上場日")
    public_offering: float = Field(description="公募")
    evaluation: str = Field(description="評価")
    initial_price: float = Field(description="初値")

    @staticmethod
    def from_dict(data: dict) -> "DecodeHtmlPageStockIpo":
        manager = ""
        stock_listing_at = ""
        public_offering = 0
        evaluation = ""
        initial_price = 0
        if "主幹" in data:
            manager = data["主幹"]
        elif "manager" in data:
            manager = data["manager"]

        if "上場" in data:
            stock_listing_at = data["上場"]
        elif "stock_listing_at" in data:
            stock_listing_at = data["stock_listing_at"]

        if "公募" in data:
            public_offering = data["公募"]
        elif "public_offering" in data:
            public_offering = data["public_offering"]

        if "評価" in data:
            evaluation = data["評価"]
        elif "evaluation" in data:
            evaluation = data["evaluation"]

        if "初値" in data:
            initial_price = data["初値"]
        elif "initial_price" in data:
            initial_price = data["initial_price"]

        return DecodeHtmlPageStockIpo(
            code=data["code"],
            manager=manager,
            stock_listing_at=stock_listing_at,
            public_offering=convert_float(public_offering),
            evaluation=evaluation,
            initial_price=convert_float(initial_price),
        )

    def to_dict(self) -> dict:
        return self.model_dump()
